#include <stdlib.h>
#include <stdio.h>

#include "list.h"
#include "graph.h"

static int
_graph_cmp(const void *a, const void *b)
{
	return ((a == b) ? 0 : 1);
}

static void
_graph_free_edges(void *data)
{
	graph_node *np;
	graph_edge *ep;
	list *cnt;

	np = (graph_node *) data;
	cnt = np->in_edges;

	while (cnt != NULL) {
		ep = (graph_edge *) cnt->data;

		if (ep != NULL) {
			list_del(&(ep->src->out_edges), ep);
		}

		free(ep);
		cnt = cnt->next;
	}

	list_free(&(np->in_edges));
	cnt = np->out_edges;

	while (cnt != NULL) {
		ep = (graph_edge *) cnt->data;

		if (ep != NULL) {
			list_del(&(ep->dst->in_edges), ep);
		}

		free(ep);
		cnt = cnt->next;
	}

	list_free(&(np->out_edges));
}

static void
_graph_free_edges_list(list *lp, void *args)
{
	_graph_free_edges(lp->data);
}

static void
_graph_free_nodes(list *lp)
{
	list *next;

	next = lp;
	while (next != NULL) {
		next = lp->next;

		free(lp->data);
		free(lp);

		lp = next;
	}
}

graph *
graph_new(void)
{
	graph *new;

	new = malloc(sizeof(*new));
	if (new != NULL) {
		new->nodes = NULL;
	}

	return new;
}

void
graph_free(graph **gp)
{
	list_apply((*gp)->nodes, _graph_free_edges_list, NULL);
	_graph_free_nodes((*gp)->nodes);

	free(*gp);
	(*gp) = NULL;
}

graph_node *
graph_new_node(void)
{
	graph_node *new;

	new = malloc(sizeof(*new));
	if (new != NULL) {
		new->in_edges = NULL;
		new->out_edges = NULL;
		new->data = NULL;
	}

	return new;
}

void *
graph_set_node_data(graph_node *np, void *data)
{
	void *old;

	old = np->data;
	np->data = data;

	return old;
}

int
graph_add_node(graph *gp, graph_node *np)
{
	list *new;

	if ((gp == NULL) || (np == NULL)) {
		return GRAPH_INVALID_ARGS;
	}

	new = list_new();
	if (new == NULL) {
		return GRAPH_NO_MEMORY;
	}

	new->data = np;
	new->next = gp->nodes;
	gp->nodes = new;

	return GRAPH_OK;
}

graph_node *
graph_lookup_node(graph *gp, graph_node *np, node_cmp cmp, int insert)
{
	list *tmp;
	graph_node *res;

	if ((gp == NULL) || (np == NULL) || (cmp == NULL)) {
		return NULL;
	}

	tmp = gp->nodes;

	while (tmp != NULL) {
		res = tmp->data;

		if ((cmp)(res->data, np->data) == 0) {
			return res;
		}

		tmp = tmp->next;
	}

	if (insert != 0) {
		if (graph_add_node(gp, np) == GRAPH_OK) {
			return np;
		}
	}

	return NULL;
}

int
graph_del_node(graph *gp, graph_node *np)
{
	list tmp, *res;

	if ((gp == NULL) || (np == NULL)) {
		return GRAPH_INVALID_ARGS;
	}

	tmp.data = np;
	res = list_lookup(&(gp->nodes), &tmp, _graph_cmp, 0);
	if (res == NULL) {
		return GRAPH_INVALID_ARGS;
	}

	list_del(&(gp->nodes), np);
	_graph_free_edges(np);
	free(np);

	return GRAPH_OK;
}

graph_edge *
graph_new_edge(void)
{
	graph_edge *new;

	new = malloc(sizeof(*new));
	if (new != NULL) {
		new->src = NULL;
		new->dst = NULL;
		new->data = NULL;
	}

	return new;
}

graph_node *
graph_set_edge_src(graph_edge *ep, graph_node *np)
{
	graph_node *old;

	old = ep->src;
	ep->src = np;

	return old;
}

graph_node *
graph_set_edge_dst(graph_edge *ep, graph_node *np)
{
	graph_node *old;

	old = ep->dst;
	ep->dst = np;

	return old;
}

void *
graph_set_edge_data(graph_edge *ep, void *data)
{
	void *old;

	old = ep->data;
	ep->data = data;

	return old;
}

int
graph_add_edge(graph_edge *ep)
{
	graph_node *dst, *src;
	list *in, *out;

	src = ep->src;
	dst = ep->dst;

	if ((src == NULL) || (dst == NULL) || (ep == NULL)) {
		return GRAPH_INVALID_ARGS;
	}

	if (graph_lookup_edge(ep, _graph_cmp, 0) != NULL) {
		return GRAPH_INVALID_ARGS;
	}

	in = list_new();
	if (in == NULL) {
		goto err_in;
	}

	out = list_new();
	if (out == NULL) {
		goto err_out;
	}

	in->next = dst->in_edges;
	in->data = ep;
	dst->in_edges = in;

	out->next = src->out_edges;
	out->data = ep;
	src->out_edges = out;

	goto normal;
err_out:
	free(in);
err_in:
	return GRAPH_NO_MEMORY;
normal:
	return GRAPH_OK;
}

graph_edge *
graph_lookup_edge(graph_edge *ep, edge_cmp cmp, int insert)
{
	list *tmp, *curr;
	graph_edge *res;

	if ((ep->dst == NULL) || (ep->src == NULL)) {
		return NULL;
	}

	curr = ep->src->out_edges;
	while (curr != NULL) {

		res = curr->data;
		if ((cmp)(res->data, ep->data) == 0) {
			if (res->dst == ep->dst) {
				list hint;

				hint.data = res;

				tmp = list_lookup(&(ep->dst->in_edges), &hint, _graph_cmp, 0);
				if (tmp != NULL) {
					return res;
				}
			}
		}

		curr = curr->next;
	}

	if (insert != 0) {
		if (graph_add_edge(ep) == GRAPH_OK) {
			return ep;
		}
	}

	return NULL;
}

void
graph_del_edge(graph_edge *ep)
{
	if (graph_lookup_edge(ep, _graph_cmp, 0) != NULL) {
		list_del(&(ep->src->out_edges), ep);
		list_del(&(ep->dst->in_edges), ep);
		free(ep);
	}
}

